import os
from typing import Optional

import requests  # type: ignore
from pydantic import BaseModel

from .error import (
    APIError,
    AuthorizationError,
    HTTPError,
    RateLimitError,
    UnauthenticatedRateLimitError,
)

api_base_url_v1 = "https://api.elevenlabs.io/v1"


class API(BaseModel):
    class Config:
        # Parse enum to strings when converting to dict
        use_enum_values = True
        # Validate fields when setting manually
        validate_assignment = True

    @staticmethod
    def request(url: str, method: str, api_key: Optional[str] = None, **kwargs):
        api_key = api_key or os.environ.get("ELEVEN_API_KEY")
        headers = {"xi-api-key": api_key}

        if method == "get":
            response = requests.get(url, headers=headers, **kwargs)
        elif method == "post":
            response = requests.post(url, headers=headers, **kwargs)
        else:
            raise ValueError(f"Invalid request method {method}")

        status_code = response.status_code

        if status_code == 200:
            return response

        error = HTTPError(response)

        if status_code == 401:
            if error.status == "quota_exceeded":
                if api_key is None:
                    raise UnauthenticatedRateLimitError(error)
                else:
                    raise RateLimitError(error)
            elif error.status == "needs_authorization":
                raise AuthorizationError(error)

        raise APIError(error)

    @staticmethod
    def get(url: str, *args, **kwargs):
        return API.request(url, method="get", *args, **kwargs)  # type: ignore

    @staticmethod
    def post(url: str, *args, **kwargs):
        return API.request(url, method="post", *args, **kwargs)  # type: ignore



from __future__ import annotations

from collections.abc import Iterator

from .base import API, api_base_url_v1
from .voice import Voice


class TTS(API):
    @staticmethod
    def generate(text: str, voice: Voice) -> bytes:
        url = f"{api_base_url_v1}/text-to-speech/{voice.voice_id}"
        data = dict(
            text=text, voice_settings=voice.settings.dict() if voice.settings else None
        )  # type: ignore
        response = API.post(url, json=data)
        return response.content

    @staticmethod
    def generate_stream(
        text: str, voice: Voice, stream_chunk_size: int = 2048
    ) -> Iterator[bytes]:
        url = f"{api_base_url_v1}/text-to-speech/{voice.voice_id}/stream"
        data = dict(
            text=text, voice_settings=voice.settings.dict() if voice.settings else None
        )  # type: ignore
        response = API.post(url, json=data, stream=True)
        for chunk in response.iter_content(chunk_size=stream_chunk_size):
            if chunk:
                yield chunk



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Tuple

from pydantic import Field, root_validator, validator

from .base import API, api_base_url_v1
from .error import APIError


class UnauthorizedVoiceCloningError(APIError):
    pass


class VoiceSettings(API):
    stability: float = Field(..., ge=0.0, le=1.0)
    similarity_boost: float = Field(..., ge=0.0, le=1.0)


class VoiceClone(API):
    name: str = Field(..., min_length=1, max_length=100)
    description: str = ""
    files: List[str] = Field(..., min_items=1, max_items=25)
    labels: Optional[dict[str, str]]
    _files_tuple: Optional[List[Tuple]] = None

    @root_validator
    def computed_files_tuple(cls, values) -> List[str]:
        files_tuple = []
        for filepath in values["files"]:
            b = open(filepath, "rb")
            file_tuple = ("files", (f"{Path(filepath).stem}_{id(b)}", b, "audio/mpeg"))
            files_tuple.append(file_tuple)
        values["_files_tuple"] = files_tuple
        return values


class Voice(API):
    voice_id: str
    name: Optional[str]
    category: Optional[str]
    settings: Optional[VoiceSettings]

    @classmethod
    def from_id(cls, voice_id: str):
        url = f"{api_base_url_v1}/voices/{voice_id}?with_settings=true"
        return cls(**cls.get(url).json())

    @classmethod
    def from_clone(cls, voice_clone: VoiceClone) -> Voice:
        url = f"{api_base_url_v1}/voices/add"
        data = voice_clone.dict()
        data["lables"] = str(data.pop("labels"))
        del data["files"]
        files = data.pop("_files_tuple")
        try:
            voice_id = API.post(url, data=data, files=files).json()["voice_id"]
        except APIError as e:
            if e.http_error.status == "can_not_use_instant_voice_cloning":
                raise UnauthorizedVoiceCloningError(e.http_error)
            raise
        return voice_id

    @validator("settings")
    def computed_settings(cls, v: VoiceSettings, values) -> VoiceSettings:
        url = f"{api_base_url_v1}/voices/{values['voice_id']}/settings"
        return v if v else VoiceSettings(**cls.get(url).json())


class Voices(API):
    voices: List[Voice]

    @classmethod
    def from_api(cls):
        url = f"{api_base_url_v1}/voices"
        response = cls.get(url).json()
        return cls(**response)

    def add_clone(self, voice_clone: VoiceClone) -> Voice:
        pass

    def __getitem__(self, idx: int) -> Voice:
        return self.voices[idx]

    def __iter__(self):
        return iter(self.voices)



import os
import re
from collections.abc import Iterator
from typing import List, Optional, Union

from .api import TTS, Voice, VoiceClone, Voices, VoiceSettings


def set_api_key(api_key: str) -> None:
    os.environ["ELEVEN_API_KEY"] = api_key


def get_api_key() -> Optional[str]:
    return os.environ.get("ELEVEN_API_KEY")


# Save default voices to avoid querying the API for unathorized users
VOICES_CACHE = [
    Voice(
        voice_id="21m00Tcm4TlvDq8ikWAM",
        name="Rachel",
        category="premade",
        settings=VoiceSettings(stability=0.75, similarity_boost=0.75),
    ),
    Voice(
        voice_id="AZnzlk1XvdvUeBnXmlld",
        name="Domi",
        category="premade",
        settings=VoiceSettings(stability=0.1, similarity_boost=0.75),
    ),
    Voice(
        voice_id="EXAVITQu4vr4xnSDxMaL",
        name="Bella",
        category="premade",
        settings=VoiceSettings(stability=0.245, similarity_boost=0.75),
    ),
    Voice(
        voice_id="ErXwobaYiN019PkySvjV",
        name="Antoni",
        category="premade",
        settings=VoiceSettings(stability=0.195, similarity_boost=0.75),
    ),
    Voice(
        voice_id="MF3mGyEYCl7XYWbV9V6O",
        name="Elli",
        category="premade",
        settings=VoiceSettings(stability=0.755, similarity_boost=0.75),
    ),
    Voice(
        voice_id="TxGEqnHWrfWFTfGW9XjX",
        name="Josh",
        category="premade",
        settings=VoiceSettings(stability=0.15, similarity_boost=0.51),
    ),
    Voice(
        voice_id="VR6AewLTigWG4xSOukaG",
        name="Arnold",
        category="premade",
        settings=VoiceSettings(stability=0.15, similarity_boost=0.75),
    ),
    Voice(
        voice_id="pNInz6obpgDQGcFmaJgB",
        name="Adam",
        category="premade",
        settings=VoiceSettings(stability=0.2, similarity_boost=0.75),
    ),
    Voice(
        voice_id="yoZ06aMxZJJ28mfd3POQ",
        name="Sam",
        category="premade",
        settings=VoiceSettings(stability=0.25, similarity_boost=0.75),
    ),
]


def voices(api_key: Optional[str] = None) -> List[Voice]:
    """Lists all voices in the API, if authenticated for the current user"""
    if api_key:
        set_api_key(api_key)
    api_key = get_api_key()
    global VOICES_CACHE
    VOICES_CACHE = Voices.from_api() if api_key else VOICES_CACHE
    return VOICES_CACHE


def clone(**kwargs) -> Voice:
    return Voice.from_clone(VoiceClone(**kwargs))


def is_voice_id(val: str) -> bool:
    return bool(re.match(r"^[a-zA-Z0-9]{20}$", val))


def generate(
    text: str,
    api_key: Optional[str] = None,
    voice: Union[str, Voice] = VOICES_CACHE[2],  # Bella
    stream: bool = False,
    stream_chunk_size: int = 2048,
) -> Union[bytes, Iterator[bytes]]:
    if api_key:
        set_api_key(api_key)

    if isinstance(voice, str):
        voice_str = voice
        # If voice is valid voice_id, use it
        if is_voice_id(voice):
            voice = Voice(voice_id=voice)
        # Otherwise, search voice by name
        else:
            # Check if voice is in cache
            voice = next((v for v in VOICES_CACHE if v.name == voice), None)  # type: ignore # noqa E501
            # If not, query API
            voice = next((v for v in voices() if v.name == voice), None) if not voice else voice  # type: ignore # noqa E501
        # Raise error if voice not found
        if not voice:
            raise ValueError(f"Voice '{voice_str}' not found.")

    assert isinstance(voice, Voice)

    if stream:
        return TTS.generate_stream(text, voice, stream_chunk_size)
    else:
        return TTS.generate(text, voice)



from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import root_validator

from .base import API, api_base_url_v1
from .voice import VoiceSettings


class HistoryItem(API):
    history_item_id: str
    request_id: Optional[str]
    voice_id: str
    text: str
    date: Optional[datetime]
    date_unix: int
    character_count_change_from: int
    character_count_change_to: int
    character_count_change: Optional[int]
    content_type: str
    settings: Optional[VoiceSettings]
    feedback: Optional[str]

    @root_validator
    def computed(cls, values):
        # Compute character count field
        change_from = values["character_count_change_from"]
        change_to = values["character_count_change_to"]
        values["character_count_change"] = change_to - change_from
        # Compute datetime field
        values["date"] = datetime.utcfromtimestamp(values["date_unix"])
        return values


class History(API):
    history: List[HistoryItem]

    @classmethod
    def from_api(cls) -> History:
        url = f"{api_base_url_v1}/history"
        response = cls.get(url).json()
        return cls(**response)

    def __getitem__(self, idx: int) -> HistoryItem:
        return self.history[idx]

    def __iter__(self):
        return iter(self.history)



from __future__ import annotations

from typing import List

from .base import API, api_base_url_v1


class Models(API):
    model_id: str
    display_name: str


class Subscription(API):
    character_count: int
    character_limit: int
    available_models: List[Models]
    status: str


class User(API):
    subscription: Subscription

    @classmethod
    def from_api(cls) -> User:
        url = f"{api_base_url_v1}/user"
        response = cls.get(url).json()
        return cls(**response)



import json
from typing import Optional

import requests  # type: ignore


class HTTPError:
    status: str
    message: str
    additional_info: Optional[dict] = None

    def __init__(self, response: requests.Response):
        detail = json.loads(response.text)["detail"]
        self.message = detail["message"]
        self.status = detail["status"]
        self.additional_info = detail.get("additional_info", None)


class APIError(Exception):
    message: Optional[str] = None

    def __init__(self, http_error: HTTPError):
        self.message = self.message or http_error.message
        self.http_error = http_error
        super().__init__(self.message)


class AuthorizationError(APIError):
    message: str = "This endpoint requires a valid API key, but none was found."


class RateLimitError(APIError):
    pass


class UnauthenticatedRateLimitError(RateLimitError):
    message: str = (
        "Thanks for trying out our speech synthesis! You have reached the limit of"
        " unauthenticated requests. You can continue, for free, by setting a valid API"
        " key."
    )



import os
import shutil
import subprocess
from collections.abc import Iterator
from pathlib import Path


def is_installed(lib_name: str) -> bool:
    lib = shutil.which(lib_name)
    if lib is None:
        return False
    global_path = Path(lib)
    # else check if path is valid and has the correct access rights
    return global_path.exists() and os.access(global_path, os.X_OK)


def play(audio: bytes, notebook: bool = False) -> None:
    if notebook:
        from IPython.display import Audio, display

        display(Audio(audio, rate=44100, autoplay=True))
    else:
        if not is_installed("ffplay"):
            raise ValueError("ffplay from ffmpeg not found, necessary to play audio.")
        args = ["ffplay", "-autoexit", "-", "-nodisp"]
        proc = subprocess.Popen(
            args=args,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = proc.communicate(input=audio)
        proc.poll()


def save(audio: bytes, filename: str) -> None:
    with open(filename, "wb") as f:
        f.write(audio)


def stream(audio_stream: Iterator[bytes]) -> None:
    if not is_installed("mpv"):
        raise ValueError("mpv not found, necessary to stream audio.")

    mpv_command = ["mpv", "--no-cache", "--no-terminal", "--", "fd://0"]
    mpv_process = subprocess.Popen(
        mpv_command,
        stdin=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    for chunk in audio_stream:
        if chunk is not None:
            mpv_process.stdin.write(chunk)  # type: ignore
            mpv_process.stdin.flush()  # type: ignore

    if mpv_process.stdin:
        mpv_process.stdin.close()
    mpv_process.wait()



from setuptools import find_packages, setup

setup(
    name="elevenlabs",
    packages=find_packages(exclude=[]),
    version="0.2.7",
    description="The official elevenlabs python package.",
    long_description_content_type="text/markdown",
    author="Elevenlabs",
    url="https://github.com/elevenlabs/elevenlabs-python",
    keywords=["artificial intelligence", "deep learning"],
    install_requires=[
        "pydantic>=1.10",
        "ipython>=7.0",
        "requests>=2.20",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
)
